"use strict";

require("core-js/modules/es.array.concat");

require("core-js/modules/es.array.index-of");

require("core-js/modules/es.array.slice");

require("core-js/modules/es.promise");

require("core-js/modules/es.string.match");

require("core-js/modules/es.string.trim");

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

const _require = require('../utilities'),
      delay = _require.delay;

const systemExecutor = require('./executor').systemExecutor;
/**
 * @param commandExecutor   A function that returns a promise to execute a given command.
 * @constructor
 */


class Connect {
  constructor(commandExecutor) {
    this.commandExecutor = commandExecutor || systemExecutor;
  }

  _execWiFiCommand(cmdArgs) {
    return this._exec(['netsh', 'wlan'].concat(cmdArgs));
  }

  _exec(cmdArgs) {
    return this.commandExecutor(cmdArgs);
  }
  /**
   * Retrieves the profile name of the currently connected network.
   * @returns {Promise.<String>}  The profile name of the currently connected network, or undefined if no connection.
   */


  current() {
    var _this = this;

    return _asyncToGenerator(function* () {
      const iface = yield _this.currentInterface();
      return iface ? iface.profile : undefined;
    })();
  }
  /**
   * Determine the current network interface.
   * @return {Promise.<Object>} the current network interface object
   */


  currentInterface() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      const output = yield _this2._execWiFiCommand(['show', 'interfaces']);

      const lines = _this2._stringToLines(output);

      let iface = _this2._currentFromInterfaces(lines);

      if (iface && !iface['profile']) {
        iface = null;
      }

      return iface;
    })();
  }
  /**
   * Connect the wifi interface to the given access point with the named profile.
   * If the profile already exists, it is used. Otherwise a new profile for an open AP is created.
   */


  connect(profile) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const iface = yield _this3.currentInterface();
      const ifaceName = yield _this3._checkHasInterface(iface);
      const profiles = yield _this3.listProfiles(ifaceName);
      yield _this3._createProfileIfNeeded(profile, ifaceName, profiles);
      return _this3._connectProfile(profile, ifaceName);
    })();
  }

  _connectProfile(profile, interfaceName) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      yield _this4._execWiFiCommand(['connect', `name=${profile}`, `interface=${interfaceName}`]);
      yield _this4.waitForConnected(profile, interfaceName, 20, 500);
      return {
        ssid: profile
      };
    })();
  }

  waitForConnected(profile, interfaceName, count, retryPeriod) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      const ssid = yield _this5.current();

      if (ssid !== profile) {
        if (--count <= 0) {
          throw new Error(`unable to connect to network ${profile}`);
        }

        yield delay(retryPeriod);
        return _this5.waitForConnected(profile, interfaceName, count, retryPeriod);
      }

      return ssid;
    })();
  }
  /**
   * Create the profile if it doesn't already exist.
   * @param profile           The name of the profile to create (and the SSID of the open network to connect to.)
   * @param interfaceName     The interface to create the profile on.
   * @param profiles          The current list of profiles.
   * @return the profile name or a promise to create the profile, resolving to the profile name
   * @private
   */


  _createProfileIfNeeded(profile, interfaceName, profiles) {
    if (!this._profileExists(profile, profiles)) {
      return this._createProfile(profile, interfaceName);
    }

    return profile;
  }

  _profileExists(profile, profiles) {
    return profiles.indexOf(profile) >= 0;
  }
  /**
   * Creates a open AP profile so that the AP can be subsequently connected to.
   * @param profile           The name of the profile and the SSID to connect to
   * @param interfaceName     The wifi interface to register the profile with
   * @param _fs
   * @returns {*}
   * @private
   */


  _createProfile(profile, interfaceName, fs) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      fs = fs || require('fs');
      const filename = '_wifi_profile.xml';

      const content = _this6._buildProfile(profile);

      const args = ['add', 'profile', `filename=${filename}`];
      fs.writeFileSync(filename, content);

      if (interfaceName) {
        args.push(`interface=${interfaceName}`);
      }

      try {
        return _this6._execWiFiCommand(args);
      } finally {
        fs.unlinkSync(filename);
      }
    })();
  }
  /**
   * Validates that the given interface object is properly defined.
   * @param {object} iface    The object to validate
   * @throws Error if the interface is not valid
   * @private
   */


  _checkHasInterface(iface) {
    // todo - make this a programmatically identifiable error
    if (!iface || !iface.name) {
      throw Error('no Wi-Fi interface detected');
    }

    return iface.name;
  }
  /**
   * Lists all the profiles registered, either for all interfaces or for a specific interface.
   * @param {string} ifaceName    The name of the interface to list profiles for.
   * @returns {Promise.<Array.<string>>}  An array of profile names
   */


  listProfiles(ifaceName) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      const cmd = ['show', 'profiles'];

      if (ifaceName) {
        cmd.push(`interface=${ifaceName}`);
      }

      const output = yield _this7._execWiFiCommand(cmd);

      const lines = _this7._stringToLines(output);

      return _this7._parseProfiles(lines);
    })();
  }
  /**
   * Parses the output of the "show profiles" command. Profiles are "type : name"-style key-value.
   * @param lines
   * @returns {Array}
   * @private
   */


  _parseProfiles(lines) {
    const profiles = [];

    for (let i = 0; i < lines.length; i++) {
      const kv = this._keyValue(lines[i]);

      if (kv && kv.key && kv.value) {
        profiles.push(kv.value);
      }
    }

    return profiles;
  }
  /**
   * Extracts the current interface from the list of interfaces.
   * @param {Array.<string>} lines    The lines from the command output.
   * @private
   */


  _currentFromInterfaces(lines) {
    let idx = 0;
    let iface;

    while (idx < lines.length && (!iface || !iface['profile'])) {
      const data = this._extractInterface(lines, idx);

      iface = data.iface;
      idx = data.range.end;
    }

    return iface;
  }
  /**
   * Reads all the lines of info up until the end, or the next 'name', collecting the property keys and values into
   * an object keyed by 'iface'. a `range` property provides `start` and `end` for the indices of the range.
   * The end index is exclusive.
   * @param lines
   * @param index
   * @private
   */


  _extractInterface(lines, index) {
    index = index || 0;
    const result = {
      iface: {},
      range: {}
    };
    const name = 'name';
    let kv;

    for (; index < lines.length; index++) {
      kv = this._keyValue(lines[index]);

      if (kv && kv.key === name) {
        // we have the start
        result.iface[kv.key] = kv.value;
        break;
      }
    }

    result.range.start = index++;

    for (; index < lines.length; index++) {
      kv = this._keyValue(lines[index]);

      if (!kv) {
        continue;
      }

      if (kv.key === name) {
        // we have the end
        break;
      }

      if (kv.key && kv.value) {
        result.iface[kv.key] = kv.value;
      }
    }

    result.range.end = index;
    return result;
  }
  /**
   * Extract a key and value from a string like ':'
   * @param line
   * @private
   */


  _keyValue(line) {
    const colonIndex = line.indexOf(':');
    let result;

    if (colonIndex > 0) {
      const key = line.slice(0, colonIndex).trim().toLowerCase();
      const value = line.slice(colonIndex + 1).trim();
      result = {
        key: key,
        value: value
      };
    }

    return result;
  }

  _stringToLines(s) {
    return s.match(/[^\r\n]+/g) || [];
  }
  /**
   * Creates a new open profile using the given ssid.
   * @param {string} ssid The ssid of the AP to connect to. It is also the name of the profile.
   * @returns {string}    The XML descriptor of the profile.
   * @private
   */


  _buildProfile(ssid) {
    // todo - xml encode profile name
    let result = '<?xml version="1.0"?> <WLANProfile xmlns="http://www.microsoft.com/networking/WLAN/profile/v1"> <name>' + ssid + '</name> <SSIDConfig> <SSID> <name>' + ssid + '</name> </SSID> </SSIDConfig>';
    result += ' <connectionType>ESS</connectionType> <connectionMode>manual</connectionMode> <MSM> <security> <authEncryption> <authentication>open</authentication> <encryption>none</encryption> <useOneX>false</useOneX> </authEncryption> </security> </MSM>';
    result += ' </WLANProfile>';
    return result;
  }

}

function asCallback(_x, _x2) {
  return _asCallback.apply(this, arguments);
}

function _asCallback() {
  _asCallback = _asyncToGenerator(function* (promise, cb) {
    try {
      const arg = yield promise;

      try {
        cb(null, arg);
      } catch (error) {// ignore callback error
      }
    } catch (error) {
      cb(error);
    }
  });
  return _asCallback.apply(this, arguments);
}

function getCurrentNetwork(cb, connect) {
  connect = connect || new Connect();
  asCallback(connect.current(), cb);
}
/**
 *
 * @param opts
 *  - ssid property is the SSID of the network to connect to.
 *  - profileName is the name of the network profile to connect to. Defaults to ssid if not defined.
 * @param cb
 * @param connect   The Connector() instance to use. If not defined a new Connector instance will be provided.
 */


function connect(opts, cb, connect) {
  connect = connect || new Connect();
  asCallback(connect.connect(opts.ssid), cb);
}

module.exports = {
  connect: connect,
  getCurrentNetwork: getCurrentNetwork,
  asCallback: asCallback,
  Connector: Connect
};